from .utils import mesh

class Backend:
    def __init__(self, app, create_protocols=None):
        from vtk.vtkWebCore import vtkWebApplication
        self._app = app
        self._protocol = None
        self._create_protocols = create_protocols

        # Need it for id/object methods
        self._app_helper = vtkWebApplication()
        self._app_helper.SetImageEncoding(0)

    def id(self, vtk_obj):
        return str(self._app_helper.GetObjectIdMap().GetGlobalId(vtk_obj))

    def object(self, vtk_id):
        return self._app_helper().GetObjectIdMap().GetVTKObject(int(vtk_id))

    def mesh(self, dataset, field_to_keep=None, point_arrays=None, cell_arrays=None):
        return mesh(dataset, field_name=field_to_keep, point_arrays=point_arrays, cell_arrays=cell_arrays)

    def scene(self, render_window):
        return self._app.protocol_call('viewport.geometry.view.get.state', self.id(render_window), True)

    def push_image(self, render_window):
        # Disable any double render...
        render_window.GetInteractor().EnableRenderOff()
        return self._app.protocol_call('viewport.image.push', { 'view': self.id(render_window) })

    def camera(self, render_window):
        camera = render_window.GetRenderers().GetFirstRenderer().GetActiveCamera()
        return {
            'focalPoint': list(camera.GetFocalPoint()),
            'parallelProjection': camera.GetParallelProjection(),
            'parallelScale': camera.GetParallelScale(),
            'position': list(camera.GetPosition()),
            'viewAngle': camera.GetViewAngle(),
            'viewUp': list(camera.GetViewUp()),
            'centerOfRotation': list(camera.GetFocalPoint()), # no center of rotation
        }

    def set_camera(self, render_window, **kwargs):
        camera = render_window.GetRenderers().GetFirstRenderer().GetActiveCamera()
        if 'focalPoint' in kwargs:
            camera.SetFocalPoint(kwargs['focalPoint'])

        if 'parallelProjection' in kwargs:
            camera.SetParallelProjection(kwargs['parallelProjection'])

        if 'parallelScale' in kwargs:
            camera.SetParallelScale(kwargs['parallelScale'])

        if 'position' in kwargs:
            camera.SetPosition(kwargs['position'])

        if 'viewAngle' in kwargs:
            camera.SetViewAngle(kwargs['viewAngle'])

        if 'viewUp' in kwargs:
            camera.SetViewUp(kwargs['viewUp'])

    def configure_protocol(self, protocol):
        self._protocol = protocol
        from .protocols import vtkWebMouseHandler, vtkWebViewPort, vtkWebPublishImageDelivery, vtkWebLocalRendering

        # Initialize vtk application helper
        self._protocol.setSharedObject("app", self._app_helper)

        # Remote rendering - image delivery
        self._protocol.registerLinkProtocol(vtkWebMouseHandler())
        self._protocol.registerLinkProtocol(vtkWebViewPort())
        self._protocol.registerLinkProtocol(vtkWebPublishImageDelivery(decode=False))

        # Remote rendering - geometry delivery
        self._protocol.registerLinkProtocol(vtkWebLocalRendering())

        # Register custom protocols
        if self._create_protocols:
            protocols = self._create_protocols()
            for p in protocols:
                self._protocol.registerLinkProtocol(p)

        # Register active view/source/...
        for key in self._app.active_objects:
            self._app_helper.GetObjectIdMap().SetActiveObject(key, self._app.active_objects[key])


import os
import base64
import mimetypes

mimetypes.init()

def abs_path(file_path, root=None):
    basepath = os.getcwd()
    if root:
        if os.path.isfile(root):
            basepath = os.path.dirname(root)
        elif os.path.isdir(root):
            basepath = root

    full_path = file_path
    if not os.path.isabs(full_path):
        full_path = os.path.join(basepath, file_path)

    return os.path.abspath(full_path)


def read_file_as_txt(file_path, root=None):
    full_path = abs_path(file_path, root)
    with open(full_path) as txt_file:
        return str(txt_file.read())

def to_mime(file_path):
    return mimetypes.guess_type(file_path)[0]

def read_file_as_base64_url(file_path, root=None):
    full_path = abs_path(file_path, root)
    with open(full_path, 'rb') as bin_file:
        encoded = base64.b64encode(bin_file.read()).decode("ascii")
        mime = to_mime(file_path)
        return f'data:{mime};base64,{encoded}'

def validate(txt, method, root):
    if not txt:
        return False

    if len(txt) < 256:
        full_path = abs_path(txt, root)
        if os.path.isfile(full_path):
            return method(full_path)

    return txt

# =============================================================================
# Internal classes
# =============================================================================

class CallbackSet:
    def __init__(self):
        self.data = []

    def add(self, callback):
        if callable(callback):
            found = False
            for c in self.data:
                if c == callback:
                    return
            self.data.append(callback)
        else:
            for c in callback:
                self.add(c)

    def clear(self):
        self.data = []

    def __iter__(self):
        return iter(self.data)

# =============================================================================
# ChangeHandler
# =============================================================================

class ChangeHandler():
    def __init__(self, app, known_state={}):
        self._app = app
        self._protocol = app.protocol
        self._known_state = known_state
        self._modified_keys = set()
        self._all_modified_keys = set()
        self._actions = []

    def modified(self, name, value):
        self._modified_keys.add(name)

    def add_action(self, ref, args=[], value=None, method=None, property=None):
        action = { 'ref': ref }
        if method:
            action['type'] = 'method'
            action['method'] = method
            action['args'] = args

        if property:
            action['type'] = 'property'
            action['property'] = property
            action['value'] = value

        self._actions.append(action)

    def __enter__(self):
        if self not in self._app._change_handlers:
            self._app._change_handlers.append(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        callbacks = CallbackSet()
        while len(self._modified_keys):
            # Gather callbacks
            for key in self._modified_keys:
                if key in self._app._change_callbacks:
                    callbacks.add(self._app._change_callbacks[key])

            self._all_modified_keys.update(self._modified_keys)
            self._modified_keys.clear()

            # Process callbacks
            for fn in callbacks:
                fn()
            callbacks.clear()

        self._app._change_handlers.remove(self)

        if self._protocol:
            modified_state = {}
            for key in self._all_modified_keys:
                if key in self._known_state and self._known_state[key] == self._app.state[key]:
                    pass
                else:
                    modified_state[key] = self._app.state[key]
            self._protocol.push_state_change(modified_state)

            if len(self._actions):
                self._protocol.push_actions(self._actions)



# To load patched methods
from ..vtk.render_window_serializer import getReferenceId

from paraview.web.protocols import ParaViewWebMouseHandler
from paraview.web.protocols import ParaViewWebViewPort
from paraview.web.protocols import ParaViewWebProtocol

# Capture the one in PV 5.9 to cover old PV versions...
# from paraview.web.protocols import ParaViewWebPublishImageDelivery

from wslink import register as exportRpc

# For: ParaViewWebPublishImageDelivery
import time
from twisted.internet import reactor
from vtkmodules.vtkWebCore import vtkWebInteractionEvent

# Fix geometry delivery
from ..vtk.render_window_serializer import serializeInstance, SynchronizationContext, getReferenceId
# from paraview.web.protocols import ParaViewWebLocalRendering

# =============================================================================
#
# Provide Image publish-based delivery mechanism
#
# =============================================================================

CAMERA_PROP_NAMES = [
  'CameraFocalPoint',
  'CameraParallelProjection',
  'CameraParallelScale',
  'CameraPosition',
  'CameraViewAngle',
  'CameraViewUp',
]

def _pushCameraLink(viewSrc, viewDstList):
  props = {}
  for name in CAMERA_PROP_NAMES:
    props[name] = getattr(viewSrc, name)
  for v in viewDstList:
    for name in CAMERA_PROP_NAMES:
      v.__setattr__(name, props[name])
  return props

class ParaViewWebPublishImageDelivery(ParaViewWebProtocol):
    def __init__(self, decode=True, **kwargs):
        ParaViewWebProtocol.__init__(self)
        self.trackingViews = {}
        self.lastStaleTime = {}
        self.staleHandlerCount = {}
        self.deltaStaleTimeBeforeRender = 0.5 # 0.5s
        self.decode = decode
        self.viewsInAnimations = []
        self.targetFrameRate = 30.0
        self.minFrameRate = 12.0
        self.maxFrameRate = 30.0

        # Camera link handling
        self.linkedViews = []
        self.linkNames = []
        self.onLinkChange = None

        # Mouse handling
        self.lastAction = 'up'
        self.activeViewId = None

    # In case some external protocol wants to monitor when link views change
    def setLinkChangeCallback(self, fn):
      self.onLinkChange = fn

    def pushRender(self, vId, ignoreAnimation = False, staleCount=0):
        if vId not in self.trackingViews:
            return

        if not self.trackingViews[vId]["enabled"]:
            return

        if not ignoreAnimation and len(self.viewsInAnimations) > 0:
            return

        if "originalSize" not in self.trackingViews[vId]:
            view = self.getView(vId)
            self.trackingViews[vId]["originalSize"] = (int(view.ViewSize[0]), int(view.ViewSize[1]))

        if "ratio" not in self.trackingViews[vId]:
            self.trackingViews[vId]["ratio"] = 1

        ratio = self.trackingViews[vId]["ratio"]
        mtime = self.trackingViews[vId]["mtime"]
        quality = self.trackingViews[vId]["quality"]
        size = [int(s * ratio) for s in self.trackingViews[vId]["originalSize"]]

        reply = self.stillRender({ "view": vId, "mtime": mtime, "quality": quality, "size": size })

        # View might have been deleted
        if not reply:
          return

        stale = reply["stale"]
        if reply["image"]:
            # depending on whether the app has encoding enabled:
            if self.decode:
                reply["image"] = base64.standard_b64decode(reply["image"]);

            reply["image"] = self.addAttachment(reply["image"]);
            reply["format"] = "jpeg"
            # save mtime for next call.
            self.trackingViews[vId]["mtime"] = reply["mtime"]
            # echo back real ID, instead of -1 for 'active'
            reply["id"] = vId
            self.publish('viewport.image.push.subscription', reply)
        if stale:
            self.lastStaleTime[vId] = time.time()
            if self.staleHandlerCount[vId] == 0:
                self.staleHandlerCount[vId] += 1
                reactor.callLater(self.deltaStaleTimeBeforeRender, lambda: self.renderStaleImage(vId, staleCount))
        else:
            self.lastStaleTime[vId] = 0


    def renderStaleImage(self, vId, staleCount=0):
        if vId in self.staleHandlerCount and self.staleHandlerCount[vId] > 0:
            self.staleHandlerCount[vId] -= 1

            if self.lastStaleTime[vId] != 0:
                delta = (time.time() - self.lastStaleTime[vId])
                # Break on staleCount otherwise linked view will always report to be stale
                # And loop forever
                if delta >= self.deltaStaleTimeBeforeRender and staleCount < 3:
                    self.pushRender(vId, False, staleCount + 1)
                elif delta < self.deltaStaleTimeBeforeRender:
                    self.staleHandlerCount[vId] += 1
                    reactor.callLater(self.deltaStaleTimeBeforeRender - delta + 0.001, lambda: self.renderStaleImage(vId, staleCount))


    def animate(self, renderAllViews=True):
        if len(self.viewsInAnimations) == 0:
            return

        nextAnimateTime = time.time() + 1.0 /  self.targetFrameRate

        # Handle the rendering of the views
        if self.activeViewId:
          self.pushRender(self.activeViewId, True)

        if renderAllViews:
          for vId in set(self.viewsInAnimations):
              if vId != self.activeViewId:
                self.pushRender(vId, True)

        nextAnimateTime -= time.time()

        if self.targetFrameRate > self.maxFrameRate:
            self.targetFrameRate = self.maxFrameRate

        if nextAnimateTime < 0:
            if nextAnimateTime < -1.0:
                self.targetFrameRate = 1
            if self.targetFrameRate > self.minFrameRate:
                self.targetFrameRate -= 1.0
            if self.activeViewId:
                # If active view, prioritize that one over the others
                # -> Divide by 2 the refresh rate of the other views
                reactor.callLater(0.001, lambda: self.animate(not renderAllViews))
            else:
                # Keep animating at the best rate we can
                reactor.callLater(0.001, lambda: self.animate())
        else:
            # We have time so let's render all
            if self.targetFrameRate < self.maxFrameRate and nextAnimateTime > 0.005:
                self.targetFrameRate += 1.0
            reactor.callLater(nextAnimateTime, lambda: self.animate())


    @exportRpc("viewport.image.animation.fps.max")
    def setMaxFrameRate(self, fps = 30):
        self.maxFrameRate = fps


    @exportRpc("viewport.image.animation.fps.get")
    def getCurrentFrameRate(self):
        return self.targetFrameRate


    @exportRpc("viewport.image.animation.start")
    def startViewAnimation(self, viewId = '-1'):
        sView = self.getView(viewId)
        realViewId = sView.GetGlobalIDAsString()

        self.viewsInAnimations.append(realViewId)
        if len(self.viewsInAnimations) == 1:
            self.animate()


    @exportRpc("viewport.image.animation.stop")
    def stopViewAnimation(self, viewId = '-1'):
        sView = self.getView(viewId)
        realViewId = sView.GetGlobalIDAsString()

        if realViewId in self.viewsInAnimations and realViewId in self.trackingViews:
            progressRendering = self.trackingViews[realViewId]['streaming']
            self.viewsInAnimations.remove(realViewId)
            if progressRendering:
                self.progressiveRender(realViewId)


    def progressiveRender(self, viewId = '-1'):
        sView = self.getView(viewId)
        realViewId = sView.GetGlobalIDAsString()

        if realViewId in self.viewsInAnimations:
            return

        if sView.GetSession().GetPendingProgress():
            reactor.callLater(self.deltaStaleTimeBeforeRender, lambda: self.progressiveRender(viewId))
        else:
            again = sView.StreamingUpdate(True)
            self.pushRender(realViewId, True)

            if again:
                reactor.callLater(0.001, lambda: self.progressiveRender(viewId))


    @exportRpc("viewport.image.push")
    def imagePush(self, options):
        view = self.getView(options["view"])
        viewId = view.GetGlobalIDAsString()

        # Make sure an image is pushed
        self.getApplication().InvalidateCache(view.SMProxy)

        self.pushRender(viewId)


    # Internal function since the reply[image] is not
    # JSON(serializable) it can not be an RPC one
    def stillRender(self, options):
        """
        RPC Callback to render a view and obtain the rendered image.
        """
        beginTime = int(round(time.time() * 1000))
        viewId = str(options["view"])
        view = self.getView(viewId)

        # If no view id provided, skip rendering
        if not viewId:
          print('No view')
          print(options)
          return None

        # Make sure request match our selected view
        if viewId != '-1' and view.GetGlobalIDAsString() != viewId:
          # We got active view rather than our request
          view = None

        # No view to render => need some cleanup
        if not view:
          # The view has been deleted, we can not render it...
          # Clean up old view state
          if viewId in self.viewsInAnimations:
            self.viewsInAnimations.remove(viewId)

          if viewId in self.trackingViews:
            del self.trackingViews[viewId]

          if viewId in self.staleHandlerCount:
            del self.staleHandlerCount[viewId]

          # the view does not exist anymore, skip rendering
          return None

        # We are in business to render our view...

        # Make sure our view size match our request
        size = view.ViewSize[0:2]
        resize = size != options.get("size", size)
        if resize:
            size = options["size"]
            if size[0] > 10 and size[1] > 10:
              view.ViewSize = size

        # Rendering options
        t = 0
        if options and "mtime" in options:
            t = options["mtime"]
        quality = 100
        if options and "quality" in options:
            quality = options["quality"]
        localTime = 0
        if options and "localTime" in options:
            localTime = options["localTime"]
        reply = {}
        app = self.getApplication()
        if t == 0:
            app.InvalidateCache(view.SMProxy)
        if self.decode:
            stillRender = app.StillRenderToString
        else:
            stillRender = app.StillRenderToBuffer
        reply_image = stillRender(view.SMProxy, t, quality)

        # Check that we are getting image size we have set if not wait until we
        # do. The render call will set the actual window size.
        tries = 10;
        while resize and list(app.GetLastStillRenderImageSize()) != size \
              and size != [0, 0] and tries > 0:
            app.InvalidateCache(view.SMProxy)
            reply_image = stillRender(view.SMProxy, t, quality)
            tries -= 1

        if not resize and options and ("clearCache" in options) and options["clearCache"]:
            app.InvalidateCache(view.SMProxy)
            reply_image = stillRender(view.SMProxy, t, quality)

        # Pack the result
        reply["stale"] = app.GetHasImagesBeingProcessed(view.SMProxy)
        reply["mtime"] = app.GetLastStillRenderToMTime()
        reply["size"] = view.ViewSize[0:2]
        reply["memsize"] = reply_image.GetDataSize() if reply_image else 0
        reply["format"] = "jpeg;base64" if self.decode else "jpeg"
        reply["global_id"] = view.GetGlobalIDAsString()
        reply["localTime"] = localTime
        if self.decode:
            reply["image"] = reply_image
        else:
            # Convert the vtkUnsignedCharArray into a bytes object, required by Autobahn websockets
            reply["image"] = memoryview(reply_image).tobytes() if reply_image else None

        endTime = int(round(time.time() * 1000))
        reply["workTime"] = (endTime - beginTime)

        return reply


    @exportRpc("viewport.image.push.observer.add")
    def addRenderObserver(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = sView.GetGlobalIDAsString()

        if not realViewId in self.trackingViews:
            observerCallback = lambda *args, **kwargs: self.pushRender(realViewId)
            startCallback = lambda *args, **kwargs: self.startViewAnimation(realViewId)
            stopCallback = lambda *args, **kwargs: self.stopViewAnimation(realViewId)
            tag = self.getApplication().AddObserver('UpdateEvent', observerCallback)
            tagStart = self.getApplication().AddObserver('StartInteractionEvent', startCallback)
            tagStop = self.getApplication().AddObserver('EndInteractionEvent', stopCallback)
            # TODO do we need self.getApplication().AddObserver('ResetActiveView', resetActiveView())
            self.trackingViews[realViewId] = { 'tags': [tag, tagStart, tagStop], 'observerCount': 1, 'mtime': 0, 'enabled': True, 'quality': 100, 'streaming': sView.GetClientSideObject().GetEnableStreaming() }
            self.staleHandlerCount[realViewId] = 0
        else:
            # There is an observer on this view already
            self.trackingViews[realViewId]['observerCount'] += 1

        self.pushRender(realViewId)
        return { 'success': True, 'viewId': realViewId }


    @exportRpc("viewport.image.push.observer.remove")
    def removeRenderObserver(self, viewId):
        sView = None
        try:
            sView = self.getView(viewId)
        except:
            print('no view with ID %s available in removeRenderObserver' % viewId)

        realViewId = sView.GetGlobalIDAsString() if sView else viewId

        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['observerCount'] -= 1

        if observerInfo['observerCount'] <= 0:
            for tag in observerInfo['tags']:
                self.getApplication().RemoveObserver(tag)
            del self.trackingViews[realViewId]
            del self.staleHandlerCount[realViewId]

        return { 'result': 'success' }


    @exportRpc("viewport.image.push.quality")
    def setViewQuality(self, viewId, quality, ratio=1, updateLinkedView=True):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = sView.GetGlobalIDAsString()
        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['quality'] = quality
        observerInfo['ratio'] = ratio

        # Handle linked view quality/ratio synch
        if updateLinkedView and realViewId in self.linkedViews:
          for vid in self.linkedViews:
            self.setViewQuality(vid, quality, ratio, False)

        # Update image size right now!
        if "originalSize" in self.trackingViews[realViewId]:
            size = [int(s * ratio) for s in self.trackingViews[realViewId]["originalSize"]]
            if 'SetSize' in sView:
                sView.SetSize(size)
            else:
                sView.ViewSize = size

        return { 'result': 'success' }


    @exportRpc("viewport.image.push.original.size")
    def setViewSize(self, viewId, width, height):
        if width < 10 or height < 10:
            return { 'result': 'size skip' }

        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = sView.GetGlobalIDAsString()
        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['originalSize'] = (int(width), int(height))

        return { 'result': 'success' }


    @exportRpc("viewport.image.push.enabled")
    def enableView(self, viewId, enabled):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = sView.GetGlobalIDAsString()
        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['enabled'] = enabled

        return { 'result': 'success' }


    @exportRpc("viewport.image.push.invalidate.cache")
    def invalidateCache(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        self.getApplication().InvalidateCache(sView.SMProxy)
        self.getApplication().InvokeEvent('UpdateEvent')
        return { 'result': 'success' }

    # -------------------------------------------------------------------------
    # View linked
    # -------------------------------------------------------------------------

    def validateViewLinks(self):
      for linkName in self.linkNames:
        simple.RemoveCameraLink(linkName)
      self.linkNames = []

      if len(self.linkedViews) > 1:
        viewList = [self.getView(vid) for vid in self.linkedViews]
        refView = viewList.pop(0)
        for view in viewList:
          linkName = '%s_%s' % (refView.GetGlobalIDAsString(), view.GetGlobalIDAsString())
          simple.AddCameraLink(refView, view, linkName)
          self.linkNames.append(linkName)

        # Synch camera state
        srcView = viewList[0]
        dstViews = viewList[1:]
        _pushCameraLink(srcView, dstViews)


    @exportRpc("viewport.view.link")
    def updateViewLink(self, viewId = None, linkState = False):
      if viewId:
        if linkState:
          self.linkedViews.append(viewId)
        else:
          try:
            self.linkedViews.remove(viewId)
          except:
            pass
        #self.validateViewLinks()

      if len(self.linkedViews) > 1:
        allViews = [self.getView(vid) for vid in self.linkedViews]
        _pushCameraLink(allViews[0], allViews[1:])

      if self.onLinkChange:
        self.onLinkChange(self.linkedViews)

      if linkState:
        self.getApplication().InvokeEvent('UpdateEvent')

      return self.linkedViews


    # -------------------------------------------------------------------------
    # Mouse handling
    # -------------------------------------------------------------------------

    @exportRpc("viewport.mouse.interaction")
    def mouseInteraction(self, event):
        """
        RPC Callback for mouse interactions.
        """
        if 'x' not in event or 'y' not in event:
            return 0

        view = self.getView(event['view'])

        if hasattr(view, 'UseInteractiveRenderingForScreenshots'):
            if event["action"] == 'down':
                view.UseInteractiveRenderingForScreenshots = 1
            elif event["action"] == 'up':
                view.UseInteractiveRenderingForScreenshots = 0

        buttons = 0
        if event["buttonLeft"]:
            buttons |= vtkWebInteractionEvent.LEFT_BUTTON
        if event["buttonMiddle"]:
            buttons |= vtkWebInteractionEvent.MIDDLE_BUTTON
        if event["buttonRight"]:
            buttons |= vtkWebInteractionEvent.RIGHT_BUTTON

        modifiers = 0
        if event["shiftKey"]:
            modifiers |= vtkWebInteractionEvent.SHIFT_KEY
        if event["ctrlKey"]:
            modifiers |= vtkWebInteractionEvent.CTRL_KEY
        if event["altKey"]:
            modifiers |= vtkWebInteractionEvent.ALT_KEY
        if event["metaKey"]:
            modifiers |= vtkWebInteractionEvent.META_KEY

        pvevent = vtkWebInteractionEvent()
        pvevent.SetButtons(buttons)
        pvevent.SetModifiers(modifiers)
        pvevent.SetX(event["x"])
        pvevent.SetY(event["y"])
        #pvevent.SetKeyCode(event["charCode"])
        retVal = self.getApplication().HandleInteractionEvent(view.SMProxy, pvevent)
        del pvevent

        self.activeViewId = view.GetGlobalIDAsString()

        if event["action"] == 'down' and self.lastAction != event["action"]:
            self.getApplication().InvokeEvent('StartInteractionEvent')

        if event["action"] == 'up' and self.lastAction != event["action"]:
            self.getApplication().InvokeEvent('EndInteractionEvent')

        #if retVal :
        #  self.getApplication().InvokeEvent('UpdateEvent')

        if self.activeViewId in self.linkedViews:
          dstViews = [self.getView(vid) for vid in self.linkedViews]
          _pushCameraLink(view, dstViews)

        self.lastAction = event["action"]

        return retVal


    @exportRpc("viewport.mouse.zoom.wheel")
    def updateZoomFromWheel(self, event):
      if 'Start' in event["type"]:
        self.getApplication().InvokeEvent('StartInteractionEvent')

      viewProxy = self.getView(event['view'])
      if viewProxy and 'spinY' in event:
        rootId = viewProxy.GetGlobalIDAsString()
        zoomFactor = 1.0 - event['spinY'] / 10.0

        if rootId in self.linkedViews:
          fp = viewProxy.CameraFocalPoint
          pos = viewProxy.CameraPosition
          delta = [fp[i] - pos[i] for i in range(3)]
          viewProxy.GetActiveCamera().Zoom(zoomFactor)
          viewProxy.UpdatePropertyInformation()
          pos2 = viewProxy.CameraPosition
          viewProxy.CameraFocalPoint = [pos2[i] + delta[i] for i in range(3)]
          dstViews = [self.getView(vid) for vid in self.linkedViews]
          _pushCameraLink(viewProxy, dstViews)
        else:
          fp = viewProxy.CameraFocalPoint
          pos = viewProxy.CameraPosition
          delta = [fp[i] - pos[i] for i in range(3)]
          viewProxy.GetActiveCamera().Zoom(zoomFactor)
          viewProxy.UpdatePropertyInformation()
          pos2 = viewProxy.CameraPosition
          viewProxy.CameraFocalPoint = [pos2[i] + delta[i] for i in range(3)]

      if 'End' in event["type"]:
        self.getApplication().InvokeEvent('EndInteractionEvent')


# =============================================================================
#
# Provide an updated geometry delivery mechanism which better matches the
# client-side rendering capability we have in vtk.js
#
# =============================================================================

class ParaViewWebLocalRendering(ParaViewWebProtocol):
    def __init__(self, **kwargs):
        super(ParaViewWebLocalRendering, self).__init__()
        self.context = SynchronizationContext()
        self.trackingViews = {}
        self.mtime = 0

    # RpcName: getArray => viewport.geometry.array.get
    @exportRpc("viewport.geometry.array.get")
    def getArray(self, dataHash, binary = False):
        if binary:
            return self.addAttachment(self.context.getCachedDataArray(dataHash, binary))
        return self.context.getCachedDataArray(dataHash, binary)

    # RpcName: addViewObserver => viewport.geometry.view.observer.add
    @exportRpc("viewport.geometry.view.observer.add")
    def addViewObserver(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = sView.GetGlobalIDAsString()

        def pushGeometry(newSubscription=False):
            simple.Render(sView)
            stateToReturn = self.getViewState(realViewId, newSubscription)
            stateToReturn['mtime'] = 0 if newSubscription else self.mtime
            self.mtime += 1
            return stateToReturn

        if not realViewId in self.trackingViews:
            observerCallback = lambda *args, **kwargs: self.publish('viewport.geometry.view.subscription', pushGeometry())
            tag = self.getApplication().AddObserver('UpdateEvent', observerCallback)
            self.trackingViews[realViewId] = { 'tags': [tag], 'observerCount': 1 }
        else:
            # There is an observer on this view already
            self.trackingViews[realViewId]['observerCount'] += 1

        self.publish('viewport.geometry.view.subscription', pushGeometry(True))
        return { 'success': True, 'viewId': realViewId }

    # RpcName: removeViewObserver => viewport.geometry.view.observer.remove
    @exportRpc("viewport.geometry.view.observer.remove")
    def removeViewObserver(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = sView.GetGlobalIDAsString()

        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['observerCount'] -= 1

        if observerInfo['observerCount'] <= 0:
            for tag in observerInfo['tags']:
                self.getApplication().RemoveObserver(tag)
            del self.trackingViews[realViewId]

        return { 'result': 'success' }

    # RpcName: getViewState => viewport.geometry.view.get.state
    @exportRpc("viewport.geometry.view.get.state")
    def getViewState(self, viewId, newSubscription=False):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        self.context.setIgnoreLastDependencies(newSubscription)

        # Get the active view and render window, use it to iterate over renderers
        renderWindow = sView.GetRenderWindow()
        renderWindowId = sView.GetGlobalIDAsString()
        viewInstance = serializeInstance(None, renderWindow, renderWindowId, self.context, 1)
        viewInstance['extra'] = {
            'vtkRefId': getReferenceId(renderWindow),
            'centerOfRotation': sView.CenterOfRotation.GetData(),
            'camera': getReferenceId(sView.GetActiveCamera())
        }

        self.context.setIgnoreLastDependencies(False)
        self.context.checkForArraysToRelease()

        if viewInstance:
            return viewInstance

        return None



from ..vtk.utils import mesh
from paraview import servermanager

def apply_default_interaction_settings():
    # ProxyManager helper
    pxm = servermanager.ProxyManager()

    # Update interaction mode
    interactionProxy = pxm.GetProxy('settings', 'RenderViewInteractionSettings')
    interactionProxy.Camera3DManipulators = [
        'Rotate', 'Pan', 'Zoom',  # -
        'Pan', 'Roll', 'Pan',     # shift
        'Zoom', 'Rotate', 'Zoom', # ctrl
    ]

    # Custom rendering settings
    renderingSettings = pxm.GetProxy('settings', 'RenderViewSettings')
    renderingSettings.LODThreshold = 102400

class Backend:
    def __init__(self, app, create_protocols=None):
        self._app = app
        self._protocol = None
        self._create_protocols = create_protocols

    def id(self, pv_proxy):
        return pv_proxy.GetGlobalIDAsString()

    def object(self, pv_id):
        try:
            pv_id = int(pv_id)
        except:
            return None
        if pv_id <= 0:
            return None
        return servermanager._getPyProxy(servermanager.ActiveConnection.Session.GetRemoteObject(pv_id))

    def mesh(self, proxy, field_to_keep=None, point_arrays=None, cell_arrays=None):
        source = proxy.GetClientSideObject()
        dataset = source.GetOutput()
        return mesh(dataset, field_name=field_to_keep, point_arrays=point_arrays, cell_arrays=cell_arrays)

    def scene(self, view_proxy):
        # flush data without requireing a render/picture
        tmp = view_proxy.SuppressRendering
        view_proxy.SuppressRendering = 1
        view_proxy.StillRender()
        view_proxy.SuppressRendering = tmp

        return self._app.protocol_call('viewport.geometry.view.get.state', self.id(view_proxy), True)

    def push_image(self, view_proxy):
        if view_proxy.EnableRenderOnInteraction:
            view_proxy.EnableRenderOnInteraction = 0
        return self._app.protocol_call('viewport.image.push', { 'view': self.id(view_proxy) })

    def camera(self, view_proxy):
        view_proxy.UpdatePropertyInformation()
        return {
            'focalPoint': list(view_proxy.CameraFocalPoint),
            'parallelProjection': view_proxy.CameraParallelProjection,
            'parallelScale': view_proxy.CameraParallelScale,
            'position': list(view_proxy.CameraPosition),
            'viewAngle': view_proxy.CameraViewAngle,
            'viewUp': list(view_proxy.CameraViewUp),
            'centerOfRotation': list(view_proxy.CenterOfRotation),
        }

    def set_camera(self, view_proxy, **kwargs):
        if 'focalPoint' in kwargs:
            view_proxy.CameraFocalPoint = kwargs['focalPoint']

        if 'parallelProjection' in kwargs:
            view_proxy.CameraParallelProjection = kwargs['parallelProjection']

        if 'parallelScale' in kwargs:
            view_proxy.CameraParallelScale = kwargs['parallelScale']

        if 'position' in kwargs:
            view_proxy.CameraPosition = kwargs['position']

        if 'viewAngle' in kwargs:
            view_proxy.CameraViewAngle = kwargs['viewAngle']

        if 'viewUp' in kwargs:
            view_proxy.CameraViewUp = kwargs['viewUp']

        if 'centerOfRotation' in kwargs:
            view_proxy.CenterOfRotation = kwargs['centerOfRotation']

    def configure_protocol(self, protocol):
        self._protocol = protocol

        from .protocols import ParaViewWebMouseHandler, ParaViewWebViewPort, ParaViewWebPublishImageDelivery, ParaViewWebLocalRendering
        from paraview.modules.vtkPVClientWeb import vtkPVWebApplication

        # Initialize vtk application helper
        protocol_app = vtkPVWebApplication()
        protocol_app.SetImageEncoding(0)
        self._protocol.setSharedObject("app", protocol_app)

        # Remote rendering - image delivery
        self._protocol.registerLinkProtocol(ParaViewWebMouseHandler())
        self._protocol.registerLinkProtocol(ParaViewWebViewPort())
        self._protocol.registerLinkProtocol(ParaViewWebPublishImageDelivery(decode=False))

        # Remote rendering - geometry delivery
        self._protocol.registerLinkProtocol(ParaViewWebLocalRendering())

        # Register custom protocols
        if self._create_protocols:
            protocols = self._create_protocols()
            for p in protocols:
                self._protocol.registerLinkProtocol(p)

        # Mimic client interactor on server side
        apply_default_interaction_settings()



def create_backend(app, name=None, create_protocols=None):
    if not name:
        from .core import Backend
        return Backend(app, create_protocols)

    if name == 'vtk':
        from .vtk import Backend
        return Backend(app, create_protocols)

    if name == 'paraview':
        from .paraview import Backend
        return Backend(app, create_protocols)


from vtkmodules.web.render_window_serializer import SynchronizationContext, serializeInstance
from vtkmodules.web.render_window_serializer import registerInstanceSerializer, wrapId

from vtkmodules.web.render_window_serializer import initializeSerializers as vtk_initializeSerializers

# -----------------------------------------------------------------------------

def getReferenceId(ref):
    if ref:
        try:
            return ref.__this__[1:17]
        except:
            idStr = str(ref)[-12:-1]
            # print('====> fallback ID %s for %s' % (idStr, ref))
            return idStr
    return '0x0'

# -----------------------------------------------------------------------------

def genericMapperSerializer(parent, mapper, mapperId, context, depth):
    # This kind of mapper requires us to get 2 items: input data and lookup
    # table
    dataObject = None
    dataObjectInstance = None
    lookupTableInstance = None
    calls = []
    dependencies = []

    if hasattr(mapper, 'GetInputDataObject'):
        mapper.GetInputAlgorithm().Update()
        dataObject = mapper.GetInputDataObject(0, 0)
    else:
        if context.debugAll:
            print('This mapper does not have GetInputDataObject method')

    if dataObject:
        dataObjectId = '%s-dataset' % mapperId
        dataObjectInstance = serializeInstance(
            mapper, dataObject, dataObjectId, context, depth + 1)

        if dataObjectInstance:
            dependencies.append(dataObjectInstance)
            calls.append(['setInputData', [wrapId(dataObjectId)]])

    lookupTable = None

    if hasattr(mapper, 'GetLookupTable'):
        lookupTable = mapper.GetLookupTable()
    else:
        if context.debugAll:
            print('This mapper does not have GetLookupTable method')

    if lookupTable:
        lookupTableId = getReferenceId(lookupTable)
        lookupTableInstance = serializeInstance(
            mapper, lookupTable, lookupTableId, context, depth + 1)
        if lookupTableInstance:
            dependencies.append(lookupTableInstance)
            calls.append(['setLookupTable', [wrapId(lookupTableId)]])

    if dataObjectInstance:
        colorArrayName = mapper.GetArrayName(
        ) if mapper.GetArrayAccessMode() == 1 else mapper.GetArrayId()
        return {
            'parent': getReferenceId(parent),
            'id': mapperId,
            'type': mapper.GetClassName(),
            'properties': {
                'resolveCoincidentTopology': mapper.GetResolveCoincidentTopology(),
                'renderTime': mapper.GetRenderTime(),
                'arrayAccessMode': mapper.GetArrayAccessMode(),
                'scalarRange': mapper.GetScalarRange(),
                'useLookupTableScalarRange': 1 if mapper.GetUseLookupTableScalarRange() else 0,
                'scalarVisibility': mapper.GetScalarVisibility(),
                'colorByArrayName': colorArrayName,
                'colorMode': mapper.GetColorMode(),
                'scalarMode': mapper.GetScalarMode(),
                'interpolateScalarsBeforeMapping': 1 if mapper.GetInterpolateScalarsBeforeMapping() else 0
            },
            'calls': calls,
            'dependencies': dependencies
        }

    return None


# -----------------------------------------------------------------------------

def rendererSerializer(parent, instance, objId, context, depth):
    dependencies = []
    viewPropIds = []
    lightsIds = []
    calls = []

    # Camera
    camera = instance.GetActiveCamera()
    cameraId = getReferenceId(camera)
    cameraInstance = serializeInstance(
        instance, camera, cameraId, context, depth + 1)
    if cameraInstance:
        dependencies.append(cameraInstance)
        calls.append(['setActiveCamera', [wrapId(cameraId)]])

    # View prop as representation containers
    viewPropCollection = instance.GetViewProps()
    for rpIdx in range(viewPropCollection.GetNumberOfItems()):
        viewProp = viewPropCollection.GetItemAsObject(rpIdx)
        viewPropId = getReferenceId(viewProp)

        viewPropInstance = serializeInstance(
            instance, viewProp, viewPropId, context, depth + 1)
        if viewPropInstance:
            dependencies.append(viewPropInstance)
            viewPropIds.append(viewPropId)

    calls += context.buildDependencyCallList('%s-props' %
                                             objId, viewPropIds, 'addViewProp', 'removeViewProp')

    # Lights
    lightCollection = instance.GetLights()
    for lightIdx in range(lightCollection.GetNumberOfItems()):
        light = lightCollection.GetItemAsObject(lightIdx)
        lightId = getReferenceId(light)

        lightInstance = serializeInstance(
            instance, light, lightId, context, depth + 1)
        if lightInstance:
            dependencies.append(lightInstance)
            lightsIds.append(lightId)

    calls += context.buildDependencyCallList('%s-lights' %
                                             objId, lightsIds, 'addLight', 'removeLight')

    if len(dependencies) > 1:
        return {
            'parent': getReferenceId(parent),
            'id': objId,
            'type': instance.GetClassName(),
            'properties': {
                'background': instance.GetBackground(),
                'background2': instance.GetBackground2(),
                'viewport': instance.GetViewport(),
                # These commented properties do not yet have real setters in vtk.js
                # 'gradientBackground': instance.GetGradientBackground(),
                # 'aspect': instance.GetAspect(),
                # 'pixelAspect': instance.GetPixelAspect(),
                # 'ambient': instance.GetAmbient(),
                'twoSidedLighting': instance.GetTwoSidedLighting(),
                'lightFollowCamera': instance.GetLightFollowCamera(),
                'layer': instance.GetLayer(),
                'preserveColorBuffer': instance.GetPreserveColorBuffer(),
                'preserveDepthBuffer': instance.GetPreserveDepthBuffer(),
                'nearClippingPlaneTolerance': instance.GetNearClippingPlaneTolerance(),
                'clippingRangeExpansion': instance.GetClippingRangeExpansion(),
                'useShadows': instance.GetUseShadows(),
                'useDepthPeeling': instance.GetUseDepthPeeling(),
                'occlusionRatio': instance.GetOcclusionRatio(),
                'maximumNumberOfPeels': instance.GetMaximumNumberOfPeels(),
                'interactive': instance.GetInteractive(),
            },
            'dependencies': dependencies,
            'calls': calls
        }

    return None

# -----------------------------------------------------------------------------
# Patch current serializer with some fix before they become mainstream
# -----------------------------------------------------------------------------

def initializeSerializers():
    vtk_initializeSerializers()

    # Mapper (update input + ignore lut)
    registerInstanceSerializer('vtkOpenGLPolyDataMapper', genericMapperSerializer)
    registerInstanceSerializer('vtkOpenGLPolyDataMapper', genericMapperSerializer)
    registerInstanceSerializer('vtkCompositePolyDataMapper2', genericMapperSerializer)
    registerInstanceSerializer('vtkDataSetMapper', genericMapperSerializer)

    # Renderer (add interactive flag)
    registerInstanceSerializer('vtkOpenGLRenderer', rendererSerializer)


# Auto update...
initializeSerializers()


try:
    # v9 and above
    from vtkmodules.util.numpy_support import vtk_to_numpy
    from vtkmodules.vtkFiltersGeometry import vtkGeometryFilter
except:
    # v8.1.2 and below
    print("Can't import vtkmodules. Falling back to importing vtk.")
    from vtk.util.numpy_support import vtk_to_numpy
    from vtk.vtkFiltersGeometry import vtkGeometryFilter

# Numpy to JS TypedArray
to_js_type = {
    'int8': 'Int8Array',
    'uint8': 'Uint8Array',
    'int16': 'Int16Array',
    'uint16': 'Uint16Array',
    'int32': 'Int32Array',
    'uint32': 'Uint32Array',
    'int64': 'Int32Array',
    'uint64': 'Uint32Array',
    'float32': 'Float32Array',
    'float64': 'Float64Array',
}

def np_encode(array, np_type=None):
    if np_type:
        return {
            'bvals': base64.b64encode(memoryview(array.astype(np_type).ravel(order="C"))).decode("utf-8"),
            'dtype': str(np_type),
            'shape': list(array.shape),
        }
    return {
        'bvals': base64.b64encode(memoryview(array.ravel(order="C"))).decode("utf-8"),
        'dtype': str(array.dtype),
        'shape': list(array.shape),
    }

def b64_encode_numpy(obj):
    # Convert 1D numpy arrays with numeric types to memoryviews with
    # datatype and shape metadata.
    if len(obj) == 0:
        return obj.tolist()

    dtype = obj.dtype
    if dtype.kind == 'f':
        return np_encode(obj)
    elif dtype.kind == 'b':
        return np_encode(obj, np.uint8)
    elif dtype.kind in ['u', 'i']:
        # Try to see if we can downsize the array
        max_value = np.amax(obj)
        min_value = np.amin(obj)
        signed = min_value < 0
        test_value = max(max_value, -min_value)
        if signed:
            if test_value < np.iinfo(np.int8):
                return np_encode(obj, np.int8)
            if test_value < np.iinfo(np.int16).max:
                return np_encode(obj, np.int16)
            if test_value < np.iinfo(np.int32).max:
                return np_encode(obj, np.int32)
        else:
            if test_value < np.iinfo(np.uint8).max:
                return np_encode(obj, np.uint8)
            if test_value < np.iinfo(np.uint16).max:
                return np_encode(obj, np.uint16)
            if test_value < np.iinfo(np.uint32).max:
                return np_encode(obj, np.uint32)

    # Convert all other numpy arrays to lists
    return obj.tolist()


def mesh(dataset, field_to_keep=None, point_arrays=None, cell_arrays=None):
    """Expect any dataset and extract its surface into a dash_vtk.Mesh state property"""
    if dataset is None:
        return None

    if point_arrays is None:
        point_arrays = []
    if cell_arrays is None:
        cell_arrays = []

    # Make sure we have a polydata to export
    polydata = None
    if dataset.IsA("vtkPolyData"):
        polydata = dataset
    else:
        extractSkinFilter = vtkGeometryFilter()
        extractSkinFilter.SetInputData(dataset)
        extractSkinFilter.Update()
        polydata = extractSkinFilter.GetOutput()

    if polydata.GetPoints() is None:
        return None

    # Extract mesh
    points = b64_encode_numpy(vtk_to_numpy(polydata.GetPoints().GetData()))
    verts = (
        b64_encode_numpy(vtk_to_numpy(polydata.GetVerts().GetData()))
        if polydata.GetVerts()
        else []
    )
    lines = (
        b64_encode_numpy(vtk_to_numpy(polydata.GetLines().GetData()))
        if polydata.GetLines()
        else []
    )
    polys = (
        b64_encode_numpy(vtk_to_numpy(polydata.GetPolys().GetData()))
        if polydata.GetPolys()
        else []
    )
    strips = (
        b64_encode_numpy(vtk_to_numpy(polydata.GetStrips().GetData()))
        if polydata.GetStrips()
        else []
    )

    # Extract field
    values = None
    js_types = "Float32Array"
    nb_comp = 1
    dataRange = [0, 1]
    location = None
    if field_to_keep is not None:
        p_array = polydata.GetPointData().GetArray(field_to_keep)
        c_array = polydata.GetCellData().GetArray(field_to_keep)

        if c_array:
            dataRange = c_array.GetRange(-1)
            nb_comp = c_array.GetNumberOfComponents()
            values = vtk_to_numpy(c_array)
            js_types = to_js_type[str(values.dtype)]
            location = "CellData"

        if p_array:
            dataRange = p_array.GetRange(-1)
            nb_comp = p_array.GetNumberOfComponents()
            values = vtk_to_numpy(p_array)
            js_types = to_js_type[str(values.dtype)]
            location = "PointData"

    # other arrays (points)
    point_data = []
    for name in point_arrays:
        array = polydata.GetPointData().GetArray(name)
        if array:
            dataRange = array.GetRange(-1)
            nb_comp = array.GetNumberOfComponents()
            values = vtk_to_numpy(array)
            js_types = to_js_type[str(values.dtype)]
            point_data.append(
                {
                    "name": name,
                    "values": b64_encode_numpy(values),
                    "numberOfComponents": nb_comp,
                    "type": js_types,
                    "location": "PointData",
                    "dataRange": dataRange,
                }
            )

    # other arrays (cells)
    cell_data = []
    for name in point_arrays:
        array = polydata.GetCellData().GetArray(name)
        if array:
            dataRange = array.GetRange(-1)
            nb_comp = array.GetNumberOfComponents()
            values = vtk_to_numpy(array)
            js_types = to_js_type[str(values.dtype)]
            cell_data.append(
                {
                    "name": name,
                    "values": b64_encode_numpy(values),
                    "numberOfComponents": nb_comp,
                    "type": js_types,
                    "location": "CellData",
                    "dataRange": dataRange,
                }
            )

    state = {
        "mesh": {"points": points,},
    }
    if len(verts):
        state["mesh"]["verts"] = verts
    if len(lines):
        state["mesh"]["lines"] = lines
    if len(polys):
        state["mesh"]["polys"] = polys
    if len(strips):
        state["mesh"]["strips"] = strips

    if values is not None:
        state.update(
            {
                "field": {
                    "name": field_to_keep,
                    "values": b64_encode_numpy(values),
                    "numberOfComponents": nb_comp,
                    "type": js_types,
                    "location": location,
                    "dataRange": dataRange,
                },
            }
        )

    if len(point_data):
        state.update({"pointArrays": point_data})
    if len(cell_data):
        state.update({"cellArrays": cell_data})

    return state


def volume(dataset):
    """Expect a vtkImageData and extract its setting for the dash_vtk.Volume state"""
    if dataset is None or not dataset.IsA("vtkImageData"):
        return None

    state = {
        "image": {
            "dimensions": dataset.GetDimensions(),
            "spacing": dataset.GetSpacing(),
            "origin": dataset.GetOrigin(),
        },
    }

    # Capture image orientation if any
    if hasattr(dataset, "GetDirectionMatrix"):
        matrix = dataset.GetDirectionMatrix()
        js_mat = []
        for j in range(3):
            for i in range(3):
                js_mat.append(matrix.GetElement(i, j))

        state["image"]["direction"] = js_mat

    scalars = dataset.GetPointData().GetScalars()

    if scalars is not None:
        values = vtk_to_numpy(scalars)
        js_types = to_js_type[str(values.dtype)]
        state["field"] = {
            "name": scalars.GetName(),
            "numberOfComponents": scalars.GetNumberOfComponents(),
            "dataRange": scalars.GetRange(-1),
            "type": js_types,
            "values": b64_encode_numpy(values),
        }

    return state

presets = [
    "KAAMS",
    "Cool to Warm",
    "Cool to Warm (Extended)",
    "Warm to Cool",
    "Warm to Cool (Extended)",
    "Rainbow Desaturated",
    "Cold and Hot",
    "Black-Body Radiation",
    "X Ray",
    "Grayscale",
    "BkRd",
    "BkGn",
    "BkBu",
    "BkMa",
    "BkCy",
    "Black, Blue and White",
    "Black, Orange and White",
    "Linear YGB 1211g",
    "Linear Green (Gr4L)",
    "Linear Blue (8_31f)",
    "Blue to Red Rainbow",
    "Red to Blue Rainbow",
    "Rainbow Blended White",
    "Rainbow Blended Grey",
    "Rainbow Blended Black",
    "Blue to Yellow",
    "blot",
    "CIELab Blue to Red",
    "jet",
    "rainbow",
    "erdc_rainbow_bright",
    "erdc_rainbow_dark",
    "nic_CubicL",
    "nic_CubicYF",
    "gist_earth",
    "2hot",
    "erdc_red2yellow_BW",
    "erdc_marine2gold_BW",
    "erdc_blue2gold_BW",
    "erdc_sapphire2gold_BW",
    "erdc_red2purple_BW",
    "erdc_purple2pink_BW",
    "erdc_pbj_lin",
    "erdc_blue2green_muted",
    "erdc_blue2green_BW",
    "GREEN-WHITE_LINEAR",
    "erdc_green2yellow_BW",
    "blue2cyan",
    "erdc_blue2cyan_BW",
    "erdc_blue_BW",
    "BLUE-WHITE",
    "erdc_purple_BW",
    "erdc_magenta_BW",
    "magenta",
    "RED-PURPLE",
    "erdc_red_BW",
    "RED_TEMPERATURE",
    "erdc_orange_BW",
    "heated_object",
    "erdc_gold_BW",
    "erdc_brown_BW",
    "copper_Matlab",
    "pink_Matlab",
    "bone_Matlab",
    "gray_Matlab",
    "Purples",
    "Blues",
    "Greens",
    "PuBu",
    "BuPu",
    "BuGn",
    "GnBu",
    "GnBuPu",
    "BuGnYl",
    "PuRd",
    "RdPu",
    "Oranges",
    "Reds",
    "RdOr",
    "BrOrYl",
    "RdOrYl",
    "CIELab_blue2red",
    "blue2yellow",
    "erdc_blue2gold",
    "erdc_blue2yellow",
    "erdc_cyan2orange",
    "erdc_purple2green",
    "erdc_purple2green_dark",
    "coolwarm",
    "BuRd",
    "Spectral_lowBlue",
    "GnRP",
    "GYPi",
    "GnYlRd",
    "GBBr",
    "PuOr",
    "PRGn",
    "PiYG",
    "OrPu",
    "BrBG",
    "GyRd",
    "erdc_divHi_purpleGreen",
    "erdc_divHi_purpleGreen_dim",
    "erdc_divLow_icePeach",
    "erdc_divLow_purpleGreen",
    "Haze_green",
    "Haze_lime",
    "Haze",
    "Haze_cyan",
    "nic_Edge",
    "erdc_iceFire_H",
    "erdc_iceFire_L",
    "hsv",
    "hue_L60",
    "Spectrum",
    "Warm",
    "Cool",
    "Blues",
    "Wild Flower",
    "Citrus",
    "Brewer Diverging Purple-Orange (11)",
    "Brewer Diverging Purple-Orange (10)",
    "Brewer Diverging Purple-Orange (9)",
    "Brewer Diverging Purple-Orange (8)",
    "Brewer Diverging Purple-Orange (7)",
    "Brewer Diverging Purple-Orange (6)",
    "Brewer Diverging Purple-Orange (5)",
    "Brewer Diverging Purple-Orange (4)",
    "Brewer Diverging Purple-Orange (3)",
    "Brewer Diverging Spectral (11)",
    "Brewer Diverging Spectral (10)",
    "Brewer Diverging Spectral (9)",
    "Brewer Diverging Spectral (8)",
    "Brewer Diverging Spectral (7)",
    "Brewer Diverging Spectral (6)",
    "Brewer Diverging Spectral (5)",
    "Brewer Diverging Spectral (4)",
    "Brewer Diverging Spectral (3)",
    "Brewer Diverging Brown-Blue-Green (11)",
    "Brewer Diverging Brown-Blue-Green (10)",
    "Brewer Diverging Brown-Blue-Green (9)",
    "Brewer Diverging Brown-Blue-Green (8)",
    "Brewer Diverging Brown-Blue-Green (7)",
    "Brewer Diverging Brown-Blue-Green (6)",
    "Brewer Diverging Brown-Blue-Green (5)",
    "Brewer Diverging Brown-Blue-Green (4)",
    "Brewer Diverging Brown-Blue-Green (3)",
    "Brewer Sequential Blue-Green (9)",
    "Brewer Sequential Blue-Green (8)",
    "Brewer Sequential Blue-Green (7)",
    "Brewer Sequential Blue-Green (6)",
    "Brewer Sequential Blue-Green (5)",
    "Brewer Sequential Blue-Green (4)",
    "Brewer Sequential Blue-Green (3)",
    "Brewer Sequential Yellow-Orange-Brown (9)",
    "Brewer Sequential Yellow-Orange-Brown (8)",
    "Brewer Sequential Yellow-Orange-Brown (7)",
    "Brewer Sequential Yellow-Orange-Brown (6)",
    "Brewer Sequential Yellow-Orange-Brown (5)",
    "Brewer Sequential Yellow-Orange-Brown (4)",
    "Brewer Sequential Yellow-Orange-Brown (3)",
    "Brewer Sequential Blue-Purple (9)",
    "Brewer Sequential Blue-Purple (8)",
    "Brewer Sequential Blue-Purple (7)",
    "Brewer Sequential Blue-Purple (6)",
    "Brewer Sequential Blue-Purple (5)",
    "Brewer Sequential Blue-Purple (4)",
    "Brewer Sequential Blue-Purple (3)",
    "Brewer Qualitative Accent",
    "Brewer Qualitative Dark2",
    "Brewer Qualitative Set2",
    "Brewer Qualitative Pastel2",
    "Brewer Qualitative Pastel1",
    "Brewer Qualitative Set1",
    "Brewer Qualitative Paired",
    "Brewer Qualitative Set3",
    "Traffic Lights",
    "Traffic Lights For Deuteranopes",
    "Traffic Lights For Deuteranopes 2",
    "Muted Blue-Green",
    "Green-Blue Asymmetric Divergent (62Blbc)",
    "Asymmtrical Earth Tones (6_21b)",
    "Yellow 15",
    "Magma (matplotlib)",
    "Inferno (matplotlib)",
    "Plasma (matplotlib)",
    "Viridis (matplotlib)",
    "BlueObeliskElements",
]



from wslink import server
from wslink import register as exportRpc

from wslink.websocket import ServerProtocol
from wslink.websocket import LinkProtocol

from twisted.internet import reactor

# Only used for debug
import json

class CoreServer(ServerProtocol):
    authentication_token = "wslink-secret"
    app = None
    DEBUG = False

    # ---------------------------------------------------------------
    # Static methods
    # ---------------------------------------------------------------

    @staticmethod
    def add_arguments(parser):
        server.add_arguments(parser)

    @staticmethod
    def configure(args):
        CoreServer.authentication_token = args.authKey

    @staticmethod
    def start_webserver(args):
        server.start_webserver(options=args, protocol=CoreServer)

    # ---------------------------------------------------------------
    # Server
    # ---------------------------------------------------------------

    def initialize(self):  # Called by wslink
        self.rpcMethods = {}
        self.app = CoreServer.app
        self.app._backend.configure_protocol(self)
        self.updateSecret(CoreServer.authentication_token)

        if self.app.on_ready:
            self.app.on_ready()

    # ---------------------------------------------------------------

    def getRPCMethod(self, name):
        if len(self.rpcMethods) == 0:
            import inspect
            for protocolObject in self.getLinkProtocols():
                test = lambda x: inspect.ismethod(x) or inspect.isfunction(x)
                for k in inspect.getmembers(protocolObject.__class__, test):
                    proc = k[1]
                    if "_wslinkuris" in proc.__dict__:
                        uri_info = proc.__dict__["_wslinkuris"][0]
                        if "uri" in uri_info:
                            uri = uri_info["uri"]
                            self.rpcMethods[uri] = (protocolObject, proc)

        if name in self.rpcMethods:
            return self.rpcMethods[name]

    # ---------------------------------------------------------------
    # Publish
    # ---------------------------------------------------------------

    def push_state_change(self, modified_state):
        if CoreServer.DEBUG:
            print(f'=> Push state {len(json.dumps(modified_state))}')
        self.publish('topic.ws.vue.state', modified_state)

    # ---------------------------------------------------------------

    def push_actions(self, actions):
        self.publish('topic.ws.vue.actions', actions)

    # ---------------------------------------------------------------

    def push_layout(self, layout):
        self.publish('topic.ws.vue.layout', layout)

    # ---------------------------------------------------------------

    def push_routes(self, routes):
        self.publish('topic.ws.vue.routes', routes)

    # ---------------------------------------------------------------
    # RPCs
    # ---------------------------------------------------------------

    @exportRpc("ws.vue.init")
    def get_state(self):
        return self.app.get_initial_state()

    # ---------------------------------------------------------------

    @exportRpc("ws.vue.trigger")
    def trigger(self, name, args, kwargs):
        with self.app.capture_changes():
            self.app._triggers[name](*args, **kwargs)

    # ---------------------------------------------------------------

    @exportRpc("ws.vue.state.update")
    def update_state(self, changes):
        values_to_ignore = {}
        for change in changes:
            values_to_ignore[change['key']] = change['value']

        with self.app.capture_changes(values_to_ignore):
            for change in changes:
                self.app.set(change['key'], change['value'])


import os
import argparse
from .protocol import CoreServer
from .utils import ChangeHandler, read_file_as_txt, read_file_as_base64_url, validate, abs_path
from .backends import create_backend

from wslink import server

WWW_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'www')

class App:
    def __init__(self, name, root=None, backend=False, debug=False, create_protocols=None, **kwargs):
        self.name = name
        self._root = root
        self._backend = create_backend(self, backend, create_protocols)
        self._debug = debug

        self.on_ready = None
        self.active_objects = {}

        # May need to protect but aimed to be set by user
        self.vuetify = {}
        self.state = {}
        self.scripts = []
        self.styles = []
        self.vue_use = ['vuetify', 'router', 'vtk']
        self.serve = {}

        # properties
        self._favicon = None
        self._layout = '<div>Ws Vue default</div>'
        self._routes = []

        # @change
        self._change_callbacks = {}
        self._change_handlers = []
        self._server_keys = set()

        # @trigger
        self._triggers = {}

        # CLI argument handling
        self._parser = None

    # -------------------------------------------------------------------------
    # Annotations
    # -------------------------------------------------------------------------

    def change(self, *_args, **_kwargs):
        """
        Use as decorator `@app.change(key1, key2, ...)`
        """
        def register_change_callback(func):
            for name in _args:
                if name not in self._change_callbacks:
                    self._change_callbacks[name] = []
                self._change_callbacks[name].append(func)
                self._server_keys.add(name)
            return func

        return register_change_callback

    # -------------------------------------------------------------------------

    def trigger(self, name):
        """
        Use as decorator `@app.trigger(name)`
        """
        def register_trigger(func):
            self._triggers[name] = func
            return func

        return register_trigger

    # -------------------------------------------------------------------------
    # CLI handling
    # -------------------------------------------------------------------------

    @property
    def cli_parser(self):
        if self._parser:
            return self._parser

        self._parser = argparse.ArgumentParser(description="PyWebVue server By Kitware")
        CoreServer.add_arguments(self._parser)

        return self._parser

    @property
    def cli_args(self):
        return self.cli_parser.parse_known_args()[0]

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, value):
        self._layout = validate(value, read_file_as_txt, self._root)
        if self.protocol:
            self.protocol.push_layout(self._layout)

    # -------------------------------------------------------------------------

    @property
    def favicon(self):
        return self._favicon

    @favicon.setter
    def favicon(self, value):
        self._favicon = validate(value, read_file_as_base64_url, self._root)

    # -------------------------------------------------------------------------

    @property
    def routes(self):
        return self._routes

    @routes.setter
    def routes(self, value):
        self._routes = value
        if self.protocol:
            self.protocol.push_routes(self._routes)

    # -------------------------------------------------------------------------
    # API client / server
    # -------------------------------------------------------------------------

    def get(self, key):
        return self.state[key]

    # -------------------------------------------------------------------------

    def set(self, key, value):
        if key not in self.state or self.state[key] != value:
            self.state[key] = value
            for change_handler in self._change_handlers:
                change_handler.modified(key, value)

    # -------------------------------------------------------------------------

    def update(self, **kwargs):
        for change_handler in self._change_handlers:
            change_handler.add_action(**kwargs)

    # -------------------------------------------------------------------------

    def capture_changes(self, known_state={}):
        return ChangeHandler(self, known_state)

    # -------------------------------------------------------------------------
    # Helper methods
    # -------------------------------------------------------------------------

    def url(self, file_path):
        return read_file_as_base64_url(file_path, self._root)

    # -------------------------------------------------------------------------

    def txt(self, file_path):
        return read_file_as_txt(file_path, self._root)

    # -------------------------------------------------------------------------

    def id(self, obj):
        return self._backend.id(obj)

    # -------------------------------------------------------------------------

    def object(self, obj_id):
        return self._backend.object(obj_id)

    # -------------------------------------------------------------------------

    def mesh(self, mesh_obj, field_name=None):
        return self._backend.mesh(mesh_obj, field_name)

    # -------------------------------------------------------------------------

    def scene(self, view_obj):
        return self._backend.scene(view_obj)

    # -------------------------------------------------------------------------

    def push_image(self, view_obj):
        return self._backend.push_image(view_obj)

    # -------------------------------------------------------------------------

    def camera(self, view_obj):
        return self._backend.camera(view_obj)

    # -------------------------------------------------------------------------

    def set_camera(self, view_obj, **kwargs):
        return self._backend.set_camera(view_obj, **kwargs)

    # -------------------------------------------------------------------------
    # common
    # -------------------------------------------------------------------------

    def get_initial_state(self):
        state = {
            'name': self.name,
            'vuetify': self.vuetify,
            'layout': self.layout,
            'favicon': self.favicon,
            'routes': self.routes,
            'state': self.state,
            'scripts': self.scripts,
            'use': self.vue_use,
            'styles': self.styles,
            'stateListening': list(self._server_keys),
        }
        return state

    # -------------------------------------------------------------------------

    @property
    def protocol(self):
        return self._backend._protocol

    # -------------------------------------------------------------------------

    def protocol_call(self, method, *args, **kwargs):
        if self.protocol:
            pair = self.protocol.getRPCMethod(method)
            if pair:
                obj, func = pair
                return func(obj, *args, **kwargs)

    # -------------------------------------------------------------------------
    # Start server
    # -------------------------------------------------------------------------

    def run_server(self, port=None):
        CoreServer.app = self
        CoreServer.DEBUG = self._debug

        args = self.cli_args

        if port:
            args.port = port

        if not args.content:
            args.content = WWW_DIR

        if len(self.serve):
            endpoints = []
            for key in self.serve:
                endpoints.append(f'{key}={abs_path(self.serve[key], self._root)}')
            args.fsEndpoints = '|'.join(endpoints)
            print(args.fsEndpoints)

        CoreServer.configure(args)
        CoreServer.start_webserver(args)




class RemoteLocalView:
    def __init__(self, app, view, namespace='', interactive_ratio=1, interactive_quality=60, still_ratio=1, still_quality=98, ref='view'):
        self._app = app
        self._view = view
        self._view_id = app.id(view)

        # Image setup
        self.i_ratio = interactive_ratio
        self.i_quality = interactive_quality
        self.s_ratio = still_ratio
        self.s_quality = still_quality

        # init keys
        self.ref_key = namespace if namespace else ref
        self.mode_key = f'{namespace}.mode' if namespace else 'mode'
        self.scene_key = f'{namespace}.scene' if namespace else 'scene'
        self.camera_key = f'{namespace}.camera' if namespace else 'camera'
        self.animation_key = f'{namespace}.animate' if namespace else 'animate'

        # Attach annotations
        app.trigger(self.camera_key)(self.update_camera)
        app.trigger(f'{self.animation_key}.start')(self.start_animation)
        app.trigger(f'{self.animation_key}.stop')(self.stop_animation)


    def start_animation(self):
        self._app.set(self.mode_key, 'remote')
        self._app.protocol_call('viewport.image.push.quality', self._view_id, self.i_quality, self.i_ratio)
        self._app.protocol_call('viewport.image.animation.start', self._view_id)

    def stop_animation(self):
        self._app.protocol_call('viewport.image.animation.stop', self._view_id)
        self._app.protocol_call('viewport.image.push.quality', self._view_id, self.s_quality, self.s_ratio)
        self.push_geometry()
        self._app.set(self.mode_key, 'local')

    def push_geometry(self):
        self._app.set(self.scene_key, self._app.scene(self._view))

    def push_image(self):
        self._app.push_image(self._view)

    def update_camera(self, camera=None):
        if camera:
            self._app.set_camera(self._view, **camera)
            self.push_image()
        else:
            # Need to update local camera
            camera = self._app.camera(self._view)
            self._app.update(ref=self.ref_key, method='setCamera', args=[camera])



def getReferenceId(ref):
    if ref:
        try:
            return ref.__this__[1:17]
        except:
            idStr = str(ref)[-12:-1]
            # print('====> fallback ID %s for %s' % (idStr, ref))
            return idStr
    return '0x0'

class Backend:
    def __init__(self, app, create_protocols=None):
        self._app = app
        self._protocol = None
        self._create_protocols = create_protocols
        self._instance_map = {}

    def id(self, obj):
        obj_id = getReferenceId(obj)
        self._instance_map[obj_id] = obj
        return obj_id

    def object(self, obj_id):
        return self._instance_map[obj_id]

    def mesh(self, dataset, field_name=None):
        raise Exception('The "mesh" method is not suppored with the default Backend')

    def scene(self, view_of_backend):
        raise Exception('The "scene" method is not suppored with the default Backend')

    def push_image(self, view_of_backend):
        raise Exception('The "push_image" method is not suppored with the default Backend')

    def camera(self, view_of_backend):
        raise Exception('The "camera" method is not suppored with the default Backend')

    def set_camera(self, view_of_backend, **kwargs):
        raise Exception('The "set_camera" method is not suppored with the default Backend')

    def configure_protocol(self, protocol):
        self._protocol = protocol

        # Register custom protocols
        if self._create_protocols:
            protocols = self._create_protocols()
            for p in protocols:
                self._protocol.registerLinkProtocol(p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re

from setuptools import setup, find_packages

# perform the install
setup(
    name='pywebvue',
    description='Python server creating interactive vue app over WebSocket',
    long_description='Build a 3D Web application using VTK or ParaView while only writing a Python script. This project aims to streamline the existing ParaViewWeb infrastructure into something that is simpler to use by lowering the entry point complexity.',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    url=f'https://github.com/kitware/py-web-vue',
    license='BSD-3-Clause',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: JavaScript',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='Python Vue.js websocket VTK ParaView Web3D',
    packages=find_packages('src', exclude=('tests.*', 'tests')),
    package_dir={'':'src'},
    package_data={"pywebvue":["www/*", "www/*/*"]},
    install_requires=['wslink'],
)


import base64, time
import os, sys, logging, types, inspect, traceback, re, base64, time

from vtk.vtkWebCore import vtkWebApplication, vtkWebInteractionEvent

from .render_window_serializer import serializeInstance, SynchronizationContext, getReferenceId

# import Twisted reactor for later callback
from twisted.internet import reactor

from wslink import register as exportRpc
from wslink.websocket import LinkProtocol

# =============================================================================

class vtkWebProtocol(LinkProtocol):

    def getApplication(self):
        return self.getSharedObject("app")

    # no need for a setApplication anymore, but keep for compatibility
    def setApplication(self, app):
        pass

    def mapIdToObject(self, id):
        """
        Maps global-id for a vtkObject to the vtkObject instance. May return None if the
        id is not valid.
        """
        id = int(id)
        if id <= 0:
            return None
        return self.getApplication().GetObjectIdMap().GetVTKObject(id)

    def getGlobalId(self, obj):
        """
        Return the id for a given vtkObject
        """
        return self.getApplication().GetObjectIdMap().GetGlobalId(obj)

    def getView(self, vid):
        """
        Returns the view for a given view ID, if vid is None then return the
        current active view.
        :param vid: The view ID
        :type vid: str
        """
        v = self.mapIdToObject(vid)

        if not v:
            # Use active view is none provided.
            v = self.getApplication().GetObjectIdMap().GetActiveObject("VIEW")
        if not v:
            raise Exception("no view provided: %s" % vid)

        return v

    def setActiveView(self, view):
        """
        Set a vtkRenderWindow to be the active one
        """
        self.getApplication().GetObjectIdMap().SetActiveObject("VIEW", view)

# =============================================================================
#
# Handle Mouse interaction on any type of view
#
# =============================================================================

class vtkWebMouseHandler(vtkWebProtocol):

    @exportRpc("viewport.mouse.interaction")
    def mouseInteraction(self, event):
        """
        RPC Callback for mouse interactions.
        """
        view = self.getView(event['view'])

        buttons = 0
        if event["buttonLeft"]:
            buttons |= vtkWebInteractionEvent.LEFT_BUTTON
        if event["buttonMiddle"]:
            buttons |= vtkWebInteractionEvent.MIDDLE_BUTTON
        if event["buttonRight"]:
            buttons |= vtkWebInteractionEvent.RIGHT_BUTTON

        modifiers = 0
        if event["shiftKey"]:
            modifiers |= vtkWebInteractionEvent.SHIFT_KEY
        if event["ctrlKey"]:
            modifiers |= vtkWebInteractionEvent.CTRL_KEY
        if event["altKey"]:
            modifiers |= vtkWebInteractionEvent.ALT_KEY
        if event["metaKey"]:
            modifiers |= vtkWebInteractionEvent.META_KEY

        pvevent = vtkWebInteractionEvent()
        pvevent.SetButtons(buttons)
        pvevent.SetModifiers(modifiers)
        if "x" in event:
            pvevent.SetX(event["x"])
        if "y" in event:
            pvevent.SetY(event["y"])
        if "scroll" in event:
            pvevent.SetScroll(event["scroll"])
        if event["action"] == 'dblclick':
            pvevent.SetRepeatCount(2)
        #pvevent.SetKeyCode(event["charCode"])
        retVal = self.getApplication().HandleInteractionEvent(view, pvevent)
        del pvevent

        if event["action"] == 'down':
            self.getApplication().InvokeEvent('StartInteractionEvent')

        if event["action"] == 'up':
            self.getApplication().InvokeEvent('EndInteractionEvent')

        if retVal:
            self.getApplication().InvokeEvent('UpdateEvent')

        return retVal


    @exportRpc("viewport.mouse.zoom.wheel")
    def updateZoomFromWheel(self, event):
      if 'Start' in event["type"]:
        self.getApplication().InvokeEvent('StartInteractionEvent')

      renderWindow = self.getView(event['view'])
      if renderWindow and 'spinY' in event:
        zoomFactor = 1.0 - event['spinY'] / 10.0

        camera = renderWindow.GetRenderers().GetFirstRenderer().GetActiveCamera()
        fp = camera.GetFocalPoint()
        pos = camera.GetPosition()
        delta = [fp[i] - pos[i] for i in range(3)]
        camera.Zoom(zoomFactor)

        pos2 = camera.GetPosition()
        camera.SetFocalPoint([pos2[i] + delta[i] for i in range(3)])
        renderWindow.Modified()

      if 'End' in event["type"]:
        self.getApplication().InvokeEvent('EndInteractionEvent')

# =============================================================================
#
# Basic 3D Viewport API (Camera + Orientation + CenterOfRotation
#
# =============================================================================

class vtkWebViewPort(vtkWebProtocol):

    @exportRpc("viewport.camera.reset")
    def resetCamera(self, viewId):
        """
        RPC callback to reset camera.
        """
        view = self.getView(viewId)
        renderer = view.GetRenderers().GetFirstRenderer()
        renderer.ResetCamera()

        self.getApplication().InvalidateCache(view)
        self.getApplication().InvokeEvent('UpdateEvent')

        return str(self.getGlobalId(view))

    @exportRpc("viewport.axes.orientation.visibility.update")
    def updateOrientationAxesVisibility(self, viewId, showAxis):
        """
        RPC callback to show/hide OrientationAxis.
        """
        view = self.getView(viewId)
        # FIXME seb: view.OrientationAxesVisibility = (showAxis if 1 else 0);

        self.getApplication().InvalidateCache(view)
        self.getApplication().InvokeEvent('UpdateEvent')

        return str(self.getGlobalId(view))

    @exportRpc("viewport.axes.center.visibility.update")
    def updateCenterAxesVisibility(self, viewId, showAxis):
        """
        RPC callback to show/hide CenterAxesVisibility.
        """
        view = self.getView(viewId)
        # FIXME seb: view.CenterAxesVisibility = (showAxis if 1 else 0);

        self.getApplication().InvalidateCache(view)
        self.getApplication().InvokeEvent('UpdateEvent')

        return str(self.getGlobalId(view))

    @exportRpc("viewport.camera.update")
    def updateCamera(self, view_id, focal_point, view_up, position, forceUpdate = True):
        view = self.getView(view_id)

        camera = view.GetRenderers().GetFirstRenderer().GetActiveCamera()
        camera.SetFocalPoint(focal_point)
        camera.SetViewUp(view_up)
        camera.SetPosition(position)

        if forceUpdate:
            self.getApplication().InvalidateCache(view)
            self.getApplication().InvokeEvent('UpdateEvent')

# =============================================================================
#
# Provide publish-based Image delivery mechanism
#
# =============================================================================

class vtkWebPublishImageDelivery(vtkWebProtocol):
    def __init__(self, decode=True):
        super(vtkWebPublishImageDelivery, self).__init__()
        self.trackingViews = {}
        self.lastStaleTime = 0
        self.staleHandlerCount = 0
        self.deltaStaleTimeBeforeRender = 0.5 # 0.5s
        self.decode = decode
        self.viewsInAnimations = []
        self.targetFrameRate = 30.0
        self.minFrameRate = 12.0
        self.maxFrameRate = 30.0


    def pushRender(self, vId, ignoreAnimation = False):
        if vId not in self.trackingViews:
            return

        if not self.trackingViews[vId]["enabled"]:
            return

        if not ignoreAnimation and len(self.viewsInAnimations) > 0:
            return

        if "originalSize" not in self.trackingViews[vId]:
            view = self.getView(vId)
            self.trackingViews[vId]["originalSize"] = list(view.GetSize());

        if "ratio" not in self.trackingViews[vId]:
            self.trackingViews[vId]["ratio"] = 1

        ratio = self.trackingViews[vId]["ratio"]
        mtime = self.trackingViews[vId]["mtime"]
        quality = self.trackingViews[vId]["quality"]
        size = [int(s * ratio) for s in self.trackingViews[vId]["originalSize"]]

        reply = self.stillRender({ "view": vId, "mtime": mtime, "quality": quality, "size": size })
        stale = reply["stale"]
        if reply["image"]:
            # depending on whether the app has encoding enabled:
            if self.decode:
                reply["image"] = base64.standard_b64decode(reply["image"]);

            reply["image"] = self.addAttachment(reply["image"]);
            reply["format"] = "jpeg"
            # save mtime for next call.
            self.trackingViews[vId]["mtime"] = reply["mtime"]
            # echo back real ID, instead of -1 for 'active'
            reply["id"] = vId
            self.publish('viewport.image.push.subscription', reply)
        if stale:
            self.lastStaleTime = time.time()
            if self.staleHandlerCount == 0:
                self.staleHandlerCount += 1
                reactor.callLater(self.deltaStaleTimeBeforeRender, lambda: self.renderStaleImage(vId))
        else:
            self.lastStaleTime = 0


    def renderStaleImage(self, vId):
        self.staleHandlerCount -= 1

        if self.lastStaleTime != 0:
            delta = (time.time() - self.lastStaleTime)
            if delta >= self.deltaStaleTimeBeforeRender:
                self.pushRender(vId)
            else:
                self.staleHandlerCount += 1
                reactor.callLater(self.deltaStaleTimeBeforeRender - delta + 0.001, lambda: self.renderStaleImage(vId))


    def animate(self):
        if len(self.viewsInAnimations) == 0:
            return

        nextAnimateTime = time.time() + 1.0 /  self.targetFrameRate
        for vId in self.viewsInAnimations:
            self.pushRender(vId, True)

        nextAnimateTime -= time.time()

        if self.targetFrameRate > self.maxFrameRate:
            self.targetFrameRate = self.maxFrameRate

        if nextAnimateTime < 0:
            if nextAnimateTime < -1.0:
                self.targetFrameRate = 1
            if self.targetFrameRate > self.minFrameRate:
                self.targetFrameRate -= 1.0
            reactor.callLater(0.001, lambda: self.animate())
        else:
            if self.targetFrameRate < self.maxFrameRate and nextAnimateTime > 0.005:
                self.targetFrameRate += 1.0
            reactor.callLater(nextAnimateTime, lambda: self.animate())


    @exportRpc("viewport.image.animation.fps.max")
    def setMaxFrameRate(self, fps = 30):
        self.maxFrameRate = fps


    @exportRpc("viewport.image.animation.fps.get")
    def getCurrentFrameRate(self):
        return self.targetFrameRate


    @exportRpc("viewport.image.animation.start")
    def startViewAnimation(self, viewId = '-1'):
        sView = self.getView(viewId)
        realViewId = str(self.getGlobalId(sView))

        self.viewsInAnimations.append(realViewId)
        if len(self.viewsInAnimations) == 1:
            self.animate()


    @exportRpc("viewport.image.animation.stop")
    def stopViewAnimation(self, viewId = '-1'):
        sView = self.getView(viewId)
        realViewId = str(self.getGlobalId(sView))

        if realViewId in self.viewsInAnimations:
            self.viewsInAnimations.remove(realViewId)


    @exportRpc("viewport.image.push")
    def imagePush(self, options):
        sView = self.getView(options["view"])
        realViewId = str(self.getGlobalId(sView))
         # Make sure an image is pushed
        self.getApplication().InvalidateCache(sView)
        self.pushRender(realViewId)

    # Internal function since the reply[image] is not
    # JSON(serializable) it can not be an RPC one
    def stillRender(self, options):
        """
        RPC Callback to render a view and obtain the rendered image.
        """
        beginTime = int(round(time.time() * 1000))
        view = self.getView(options["view"])
        size = view.GetSize()[0:2]
        resize = size != options.get("size", size)
        if resize:
            size = options["size"]
            if size[0] > 10 and size[1] > 10:
              view.SetSize(size)
        t = 0
        if options and "mtime" in options:
            t = options["mtime"]
        quality = 100
        if options and "quality" in options:
            quality = options["quality"]
        localTime = 0
        if options and "localTime" in options:
            localTime = options["localTime"]
        reply = {}
        app = self.getApplication()
        if t == 0:
            app.InvalidateCache(view)
        if self.decode:
            stillRender = app.StillRenderToString
        else:
            stillRender = app.StillRenderToBuffer
        reply_image = stillRender(view, t, quality)

        # Check that we are getting image size we have set if not wait until we
        # do. The render call will set the actual window size.
        tries = 10;
        while resize and list(view.GetSize()) != size \
              and size != [0, 0] and tries > 0:
            app.InvalidateCache(view)
            reply_image = stillRender(view, t, quality)
            tries -= 1

        if not resize and options and ("clearCache" in options) and options["clearCache"]:
            app.InvalidateCache(view)
            reply_image = stillRender(view, t, quality)

        reply["stale"] = app.GetHasImagesBeingProcessed(view)
        reply["mtime"] = app.GetLastStillRenderToMTime()
        reply["size"] = view.GetSize()[0:2]
        reply["memsize"] = reply_image.GetDataSize() if reply_image else 0
        reply["format"] = "jpeg;base64" if self.decode else "jpeg"
        reply["global_id"] = str(self.getGlobalId(view))
        reply["localTime"] = localTime
        if self.decode:
            reply["image"] = reply_image
        else:
            # Convert the vtkUnsignedCharArray into a bytes object, required by Autobahn websockets
            reply["image"] = memoryview(reply_image).tobytes() if reply_image else None

        endTime = int(round(time.time() * 1000))
        reply["workTime"] = (endTime - beginTime)

        return reply


    @exportRpc("viewport.image.push.observer.add")
    def addRenderObserver(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = str(self.getGlobalId(sView))

        if not realViewId in self.trackingViews:
            observerCallback = lambda *args, **kwargs: self.pushRender(realViewId)
            startCallback = lambda *args, **kwargs: self.startViewAnimation(realViewId)
            stopCallback = lambda *args, **kwargs: self.stopViewAnimation(realViewId)
            tag = self.getApplication().AddObserver('UpdateEvent', observerCallback)
            tagStart = self.getApplication().AddObserver('StartInteractionEvent', startCallback)
            tagStop = self.getApplication().AddObserver('EndInteractionEvent', stopCallback)
            # TODO do we need self.getApplication().AddObserver('ResetActiveView', resetActiveView())
            self.trackingViews[realViewId] = { 'tags': [tag, tagStart, tagStop], 'observerCount': 1, 'mtime': 0, 'enabled': True, 'quality': 100 }
        else:
            # There is an observer on this view already
            self.trackingViews[realViewId]['observerCount'] += 1

        self.pushRender(realViewId)
        return { 'success': True, 'viewId': realViewId }


    @exportRpc("viewport.image.push.observer.remove")
    def removeRenderObserver(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = str(self.getGlobalId(sView))

        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['observerCount'] -= 1

        if observerInfo['observerCount'] <= 0:
            for tag in observerInfo['tags']:
                self.getApplication().RemoveObserver(tag)
            del self.trackingViews[realViewId]

        return { 'result': 'success' }


    @exportRpc("viewport.image.push.quality")
    def setViewQuality(self, viewId, quality, ratio = 1):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = str(self.getGlobalId(sView))
        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['quality'] = quality
        observerInfo['ratio'] = ratio

        # Update image size right now!
        if "originalSize" in self.trackingViews[realViewId]:
            size = [int(s * ratio) for s in self.trackingViews[realViewId]["originalSize"]]
            if hasattr(sView, 'SetSize'):
                sView.SetSize(size)
            else:
                sView.ViewSize = size

        return { 'result': 'success' }


    @exportRpc("viewport.image.push.original.size")
    def setViewSize(self, viewId, width, height):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = str(self.getGlobalId(sView))
        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['originalSize'] = [width, height]

        return { 'result': 'success' }

    @exportRpc("viewport.image.push.enabled")
    def enableView(self, viewId, enabled):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = str(self.getGlobalId(sView))
        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['enabled'] = enabled

        return { 'result': 'success' }

    @exportRpc("viewport.image.push.invalidate.cache")
    def invalidateCache(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        self.getApplication().InvalidateCache(sView)
        self.getApplication().InvokeEvent('UpdateEvent')
        return { 'result': 'success' }

# =============================================================================
#
# Provide an updated geometry delivery mechanism which better matches the
# client-side rendering capability we have in vtk.js
#
# =============================================================================

class vtkWebLocalRendering(vtkWebProtocol):
    def __init__(self, **kwargs):
        super(vtkWebLocalRendering, self).__init__()
        self.context = SynchronizationContext(True)
        self.trackingViews = {}
        self.mtime = 0

    # RpcName: getArray => viewport.geometry.array.get
    @exportRpc("viewport.geometry.array.get")
    def getArray(self, dataHash, binary = False):
        if binary:
            return self.addAttachment(self.context.getCachedDataArray(dataHash, binary))
        return self.context.getCachedDataArray(dataHash, binary)

    # RpcName: addViewObserver => viewport.geometry.view.observer.add
    @exportRpc("viewport.geometry.view.observer.add")
    def addViewObserver(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = self.getApplication().GetObjectIdMap().GetGlobalId(sView)

        def pushGeometry(newSubscription=False):
            stateToReturn = self.getViewState(realViewId, newSubscription)
            stateToReturn['mtime'] = 0 if newSubscription else self.mtime
            self.mtime += 1
            return stateToReturn

        if not realViewId in self.trackingViews:
            observerCallback = lambda *args, **kwargs: self.publish('viewport.geometry.view.subscription', pushGeometry())
            tag = self.getApplication().AddObserver('UpdateEvent', observerCallback)
            self.trackingViews[realViewId] = { 'tags': [tag], 'observerCount': 1 }
        else:
            # There is an observer on this view already
            self.trackingViews[realViewId]['observerCount'] += 1

        self.publish('viewport.geometry.view.subscription', pushGeometry(True))
        return { 'success': True, 'viewId': realViewId }

    # RpcName: removeViewObserver => viewport.geometry.view.observer.remove
    @exportRpc("viewport.geometry.view.observer.remove")
    def removeViewObserver(self, viewId):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        realViewId = self.getApplication().GetObjectIdMap().GetGlobalId(sView)

        observerInfo = None
        if realViewId in self.trackingViews:
            observerInfo = self.trackingViews[realViewId]

        if not observerInfo:
            return { 'error': 'Unable to find subscription for view %s' % realViewId }

        observerInfo['observerCount'] -= 1

        if observerInfo['observerCount'] <= 0:
            for tag in observerInfo['tags']:
                self.getApplication().RemoveObserver(tag)
            del self.trackingViews[realViewId]

        return { 'result': 'success' }

    # RpcName: getViewState => viewport.geometry.view.get.state
    @exportRpc("viewport.geometry.view.get.state")
    def getViewState(self, viewId, newSubscription=False):
        sView = self.getView(viewId)
        if not sView:
            return { 'error': 'Unable to get view with id %s' % viewId }

        self.context.setIgnoreLastDependencies(newSubscription)

        # Get the active view and render window, use it to iterate over renderers
        renderWindow = sView
        renderer = renderWindow.GetRenderers().GetFirstRenderer()
        camera = renderer.GetActiveCamera()
        renderWindowId = self.getApplication().GetObjectIdMap().GetGlobalId(sView)
        viewInstance = serializeInstance(None, renderWindow, renderWindowId, self.context, 1)
        viewInstance['extra'] = {
            'vtkRefId': getReferenceId(renderWindow),
            'centerOfRotation': camera.GetFocalPoint(),
            'camera': getReferenceId(camera)
        }

        self.context.setIgnoreLastDependencies(False)
        self.context.checkForArraysToRelease()

        if viewInstance:
            return viewInstance

        return None